from typing import Any, Optional, Union

import astropy.units as u

from ..types import Array, UnitType


def get_quantity(
    *value: Union[int, float, Array[Union[int, float]], u.Quantity],
    unit: Optional[UnitType] = None
) -> u.Quantity:
    """Convert a value to astropy Quantity.

    Parameters
    ----------
    value
        The value to be converted. If multiple values are given, they should have
        equivalent dimensions.
    unit
        The unit of the value. Should be given as a keyword argument.

    Returns
    -------
    quantity
        The value converted to Quantity.

    Raises
    ------
    UnitConversionError
        If the value has incompatible dimensions.
    ValueError
        If the values have different shapes.

    Examples
    --------
    >>> neclib.core.get_quantity(1, 2, 3, unit="m")
    <Quantity [1., 2., 3.] m>
    >>> neclib.core.get_quantity(np.array([1, 2, 3]), unit="m/s")
    <Quantity [1., 2., 3.] m / s>

    """
    if len(value) == 1:
        # Don't make 1-D array when there's only one value
        return u.Quantity(value[0], unit=unit)
    return u.Quantity(value, unit=unit)


class QuantityValidator:
    """Type validator, to force some instance variables to be Quantity."""

    # For details on validator descriptor design, see:
    # https://docs.python.org/3/howto/descriptor.html#validator-class

    def __init__(
        self, default: Union[int, float, u.Quantity] = float("nan"), *, unit: str = ""
    ) -> None:
        self.default = u.Quantity(default, unit=unit)
        self.unit = unit

    def __set_name__(self, owner: Any, name: str) -> None:
        self.private_name = "_" + name

    def __get__(self, instance: Any, owner: Any) -> u.Quantity:
        if instance is None:
            return self.default
        return getattr(instance, self.private_name, self.default)

    def __set__(self, instance: Any, value: Union[int, float, u.Quantity]) -> None:
        if not isinstance(value, u.Quantity):
            value = u.Quantity(value, self.unit)
        setattr(instance, self.private_name, value)



from typing import Any, Callable, Generic, Iterable, Iterator, Optional, TypeVar

import numpy as np

from ..types import SupportsComparison

T = TypeVar("T", bound=SupportsComparison)


class ValueRange(Generic[T]):
    """Value range bound by 2 values.

    Parameters
    ----------
    lower
        Lower bound of the range. Any type with comparison support is allowed.
    upper
        Upper bound of the range. Any type with comparison support is allowed.
    strict
        If ``True``, the value exactly equals to the bound will judged to be
        not in the range.

    Examples
    --------
    >>> valid_value = neclib.ValueRange(0, 1)
    >>> 0.5 in valid_value
    True
    >>> -1 in valid_value
    False

    You can also check lexical order of strings:

    >>> valid_str = neclib.ValueRange("aaa", "bbb")
    >>> "abc" in valid_str
    True

    The lower and upper bounds can be iterated over:

    >>> _ = [print(limits) for limits in valid_value]
    aaa
    bbb

    """

    def __init__(self, lower: T, upper: T, strict: bool = False) -> None:
        try:
            if lower > upper:
                raise ValueError("Lower bound must be smaller than upper bound.")
        except TypeError:
            raise TypeError("Bounds must support comparison.")

        self.lower, self.upper = lower, upper
        self.strict = strict

    def __contains__(self, value: Any, /) -> bool:
        if self.strict:
            return self.lower < value < self.upper
        return self.lower <= value <= self.upper

    def __iter__(self) -> Iterator[T]:
        return iter((self.lower, self.upper))

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}({self.lower}, {self.upper}, strict={self.strict})"

    def __eq__(self, other: Any, /) -> bool:
        if not isinstance(other, self.__class__):
            return False
        eq_limits = (self.lower == other.lower) and (self.upper == other.upper)
        return eq_limits and self.strict is other.strict

    @property
    def width(self) -> Optional[T]:
        """Width of the range.

        Examples
        --------
        >>> valid_value = neclib.ValueRange(0, 1)
        >>> valid_value.width
        1

        """
        try:
            return self.upper - self.lower  # type: ignore
        except TypeError:
            return None

    def contain_all(self, values: Iterable[Any]) -> bool:
        """Check if all values are in the range.

        Parameters
        ----------
        values
            Iterable object to be checked.

        Examples
        --------
        >>> valid_value = neclib.ValueRange(0, 1)
        >>> valid_value.contain_all([0.5, 1.6])
        False

        """
        values = np.asanyarray(values)
        if self.strict:
            ret = ((self.lower < values) & (values < self.upper)).all()
        else:
            ret = ((self.lower <= values) & (values <= self.upper)).all()
        return bool(ret)

    def contain_any(self, values: Iterable[Any]) -> bool:
        """Check if any value is in the range.

        Parameters
        ----------
        values
            Iterable object to be checked.

        Examples
        --------
        >>> valid_value = neclib.ValueRange(0, 1)
        >>> valid_value.contain_any([0.5, 1.6])
        True

        """
        values = np.asanyarray(values)
        if self.strict:
            ret = ((self.lower < values) & (values < self.upper)).any()
        else:
            ret = ((self.lower <= values) & (values <= self.upper)).any()
        return bool(ret)

    def map(self, func: Callable[[T], Any], /) -> "ValueRange":
        """Map a function to upper and lower bounds.

        Parameters
        ----------
        func
            Function to apply.

        Examples
        --------
        >>> valid_value = neclib.ValueRange(0, 1)
        >>> valid_value.map(lambda x: 10 * x)
        ValueRange(0, 10, strict=False)

        """
        return self.__class__(func(self.lower), func(self.upper), self.strict)



"""Inform deprecation."""

import warnings
from functools import wraps
from typing import Any, Callable, Optional


def _deprecation_message(
    name: str,
    /,
    version_since: Optional[str] = None,
    version_removed: Optional[str] = None,
    reason: Optional[str] = None,
    replacement: Optional[str] = None,
) -> str:
    message = f"The use of `{name}` is deprecated"
    if version_since:
        message += f" since version {version_since}"
    if version_removed:
        message += f" and will be removed in version {version_removed}"
    message += "."
    if reason:
        message += f" The reason for deprecation is {reason}."
    if replacement:
        message += f" Use `{replacement}` instead."
    else:
        message += " There is no replacement."
    return message


def deprecated(
    __callable: Optional[Callable[..., Any]] = None,
    *,
    version_since: Optional[str] = None,
    version_removed: Optional[str] = None,
    reason: Optional[str] = None,
    replacement: Optional[str] = None,
) -> Callable[..., Any]:
    """Decorator to mark a function as deprecated.

    Parameters
    ----------
    __callable
        A callable object to be decorated.
    version_since
        The version since which the function is deprecated.
    version_removed
        The version in which the function will be removed.
    reason
        The reason for deprecation.
    replacement
        The replacement for the deprecated function.

    Returns
    -------
    deprecated
        A callable object with deprecation warning.

    Examples
    --------
    >>> @deprecated
    ... def foo():
    ...     pass
    >>> foo()
    DeprecationWarning: The use of `foo` is deprecated. There is no replacement.

    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            message = _deprecation_message(
                func.__name__,
                version_since=version_since,
                version_removed=version_removed,
                reason=reason,
                replacement=replacement,
            )
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapped

    if __callable is None:
        return wrapper
    else:
        return wrapper(__callable)


def deprecated_namespace(*args, **kwargs):
    raise NotImplementedError



from typing import List, Tuple, Union

import numpy as np
import pandas as pd

from astropy import units as u
from astropy.coordinates import Angle
from astropy.time import Time
from matplotlib import pyplot as plt

from ...core import config
from ..convert import CoordCalculator


class OpticalPointingSpec:
    def __init__(self, time: Union[float, str], format: str) -> None:
        self.calc = CoordCalculator(config.location)
        self.now = Time(time, format=format)
        self.obsdatetime = self.now.to_datetime()

    def readlines_file(self, filename: str) -> List[str]:
        with open(filename, mode="r") as file:
            contents = file.readlines()
        return contents

    def _catalog_to_pandas(self, catalog_raw: List[str]):
        name_data = []
        ra_data = []
        dec_data = []
        multiple_data = []
        vmag_data = []
        pmra_data = []
        pmdec_data = []
        az_data = []
        el_data = []
        for line in catalog_raw:
            try:
                name = line[7:14]

                ra_raw = line[75:77] + "h" + line[77:79] + "m" + line[79:83] + "s"
                ra = Angle(ra_raw).to(u.deg)

                dec = (
                    float(line[84:86]) * u.deg
                    + float(line[86:88]) * u.arcmin
                    + float(line[88:90]) * u.arcsec
                ).to(u.deg)
                if line[83:84] == "-":
                    dec = -dec
                altaz = self.to_altaz(target=(ra, dec), frame="fk5")
                multiple = line[43:44]
                vmag = float(line[103:107])
                pmra = float(line[149:154])
                pmdec = float(line[154:160])
            except Exception:
                pass
            ra_data.append(ra.value)
            dec_data.append(dec.value)
            multiple_data.append(multiple)
            vmag_data.append(vmag)
            pmra_data.append(pmra)
            pmdec_data.append(pmdec)
            name_data.append(name)
            az_data.append(altaz.az.value)
            el_data.append(altaz.alt.value)
        data = pd.DataFrame(
            {
                "name": name_data,
                "ra": ra_data,
                "dec": dec_data,
                "pmra": pmra_data,
                "pmdec": pmdec_data,
                "az": az_data,
                "el": el_data,
                "vmag": vmag_data,
                "multiple": multiple_data,
            }
        )
        return data

    def to_altaz(self, target: Tuple[u.Quantity, u.Quantity], frame: str, time=0.0):
        if time == 0.0:
            time = self.now
        coord = self.calc.coordinate(
            lon=target[0], lat=target[1], frame=frame, time=time
        )  # TODO: Consider pressure, temperature, relative_humidity, obswl.
        altaz_coord = coord.to_apparent_altaz()
        return altaz_coord

    def _filter(
        self, catalog: pd.DataFrame, magnitude: Tuple[float, float]
    ) -> pd.DataFrame:
        az_range = config.antenna_drive_warning_limit_az
        el_range = config.antenna_drive_warning_limit_el
        filtered = catalog[
            (catalog["az"] > az_range.lower.value)
            & (catalog["az"] < az_range.upper.value)
            & (catalog["el"] > el_range.lower.value)
            & (catalog["el"] < el_range.upper.value)
            & (catalog["multiple"] == " ")
            & (catalog["pmra"] <= 1.0)
            & (catalog["pmdec"] <= 1.0)
            & (catalog["vmag"] >= magnitude[0])
            & (catalog["vmag"] <= magnitude[1])
        ]
        return filtered

    def sort(self, catalog_file: str, magnitude: Tuple[float, float]):
        az_range = config.antenna_drive_warning_limit_az

        catalog_raw = self.readlines_file(filename=catalog_file)
        catalog = self._catalog_to_pandas(catalog_raw=catalog_raw)
        catalog = self._filter(catalog, magnitude)

        sdata = catalog.sort_values("az", ignore_index=True)  # sort by az

        ddata = pd.DataFrame(index=[], columns=sdata.columns)
        elflag = 0
        azint = 100 * u.deg

        for azaz in np.arange(az_range.lower.value, az_range.upper.value, azint.value):
            ind = sdata[
                (sdata["az"] >= min(azaz, azaz + azint.value))
                & (sdata["az"] <= max(azaz, azaz + azint.value))
            ]

            ind2 = ind.sort_values("el", ignore_index=True)
            if elflag == 0:
                elflag = 1
            else:
                ind2 = ind2[::-1]
                elflag = 0
            ddata = ddata.append(ind2)
            continue
        ddata = ddata.reset_index(drop=True)

        x = ddata["az"].values.astype(np.float64)
        y = ddata["el"].values.astype(np.float64)

        show_graph = True
        if show_graph is True:
            plt.figure()
            plt.plot(x, y)
            plt.grid()
            plt.xlabel("Az")
            plt.ylabel("El")
            plt.title(
                "Optical Pointing Locus\n"
                f"obstime = {str(self.obsdatetime)}\n"
                f"star num = {str(len(ddata))}"
            )
            plt.show()

        return ddata

    def estimate_time(self, sorted_data: pd.DataFrame):
        az_speed = config.antenna.max_speed_az.value
        el_speed = config.antenna.max_speed_el.value
        time_list = []
        for i in range(len(sorted_data) - 1):
            delta_az = sorted_data["az"][i + 1] - sorted_data["az"][i]
            delta_el = sorted_data["el"][i + 1] - sorted_data["el"][i]
            if delta_az > delta_el:
                t = delta_az / az_speed
            elif delta_az < delta_el:
                t = delta_el / el_speed
            t = t + 30.0
            time_list.append(t)

        t_tot = sum(time_list)
        return t_tot



import astropy.units as u

from ...core.math import Random
from .thermometer_base import Thermometer


class ThermometerSimulator(Thermometer):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        self._rand = Random().walk(4, 0.1, -1)

    def get_temp(self, id: str) -> u.Quantity:
        return next(self._rand) * u.K

    def finalize(self) -> None:
        pass



"""General operations and object model definitions of NECLIB.

Important
---------
The implementations in this namespace should not depend on any other NECLIB subpackages
without deep consideration. Such dependencies will someday create circular imports and
make it hard to implement new features.

"""

from . import environ  # noqa: F401
from . import formatting  # noqa: F401
from . import math  # noqa: F401
from . import security  # noqa: F401
from . import types  # noqa: F401
from . import units  # noqa: F401
from .configuration import *  # noqa: F401, F403
from .data_type import *  # noqa: F401, F403
from .exceptions import *  # noqa: F401, F403
from .files import *  # noqa: F401, F403
from .inform import *  # noqa: F401, F403
from .normalization import *  # noqa: F401, F403



from typing import Callable, Union

from ...core.security import busy, sanitize
from .da_converter_base import DAConverter


class CPZ340816(DAConverter):
    """d/a converter, which can convert by 32 channels.

    Notes
    -----
    Configuration items for this device:

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default RSW1
        setting of 0. This ID would be non-zero, when multiple PCI board of same model
        are mounted on a single FA (Factory Automation) controller.

    channel : Dict[str, int]
        Human-readable channel name. The value should be mapping from human readable
        version (str) to device level identifier (int). You can assign any name to the
        channels. Defining aliases for unused channels will raise error.

    max_mv : List[int or float]
        Volatage range of this device is supplying. This setting is used for
        check weather the set value is in this range in every "voltage setting”.
        The unit of this value is mV.

    converter : str
        Function to convert setting voltage to any parameter you want.
        Supported ``x`` in function will be substituted by the setting value.
        This would be useful when setting voltage is scaled and/or
        shifted version of physical parameter.

    See defaults setting file in ``neclib/defaults/config.toml``.

    """

    Manufacturer = "Interface"
    Model = "CPZ340816"

    Identifier = "rsw_id"

    def __init__(self):
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.param_buff = {i: 0.0 for i in range(1, 17)}  # All in [V]
        self.da = pyinterface.open(3408, self.rsw_id)

    @property
    def converter(self) -> Callable[[Union[int, float]], float]:
        sanitize(self.Config.converter, "x")
        return eval(f"lambda x: {self.Config.converter}")

    def set_voltage(self, mV: float, id: str) -> None:
        ch = self.Config.channel[id]
        if ch not in self.param_buff.keys():
            raise ValueError(f"Invaild channel {ch}")
        if not self.Config.max_mv[0] < mV < self.Config.max_mv[1]:
            raise ValueError(f"Unsafe voltage {mV} mV")
        else:
            self.param_buff[ch] = self.converter(mV)

    def apply_voltage(self) -> None:
        with busy(self, "busy"):
            for i in range(0, 16):
                ch = int(list(self.param_buff.keys())[i])
                voltage = list(self.param_buff.values())[i]

                self.da.output_da(ch, voltage)

    def finalize(self) -> None:
        self.da.finalize()
        self.param_buff = {i: 0.0 for i in range(1, 17)}



import ast
from typing import List, Union


def sanitize(
    content: str, known_variables: Union[str, List[str]] = [], max_length: int = 100
) -> None:
    if len(content) > max_length:
        raise ValueError(f"Value longer than {max_length} characters is prohibited.")
    if isinstance(known_variables, str):
        known_variables = [known_variables]
    tree = ast.parse(content)
    for node in ast.walk(tree):
        if isinstance(node, ast.Name) and (node.id not in known_variables):
            raise ValueError(f"Use of unknown variable {node.id!r} is prohibited.")



import os
import shutil
from collections import defaultdict
from collections.abc import ItemsView, KeysView, ValuesView
from itertools import chain
from pathlib import Path
from typing import IO, Any, Dict, List, Union
from urllib.parse import urlparse

from astropy.coordinates import EarthLocation
from astropy.units import Quantity
from tomlkit.exceptions import ParseError

from . import environ
from .data_type import RichParameters, ValueRange
from .exceptions import NECSTConfigurationError, NECSTParameterNameError
from .files import read
from .inform import get_logger

DefaultNECSTRoot = Path.home() / ".necst"
DefaultsPath = Path(__file__).parent / ".." / "defaults"
logger = get_logger(__name__)


parsers = dict(
    location=lambda x: EarthLocation(**x),
    antenna_drive_range_az=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_range_el=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_warning_limit_az=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_warning_limit_el=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_critical_limit_az=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_critical_limit_el=lambda x: ValueRange(*map(Quantity, x)),
    antenna_pointing_parameter_path=Path,
)


class Configuration(RichParameters):

    _instances: Dict[Union[str, None], List["Configuration"]] = {}

    def __new__(cls, *args, **kwargs):
        if len(cls._instances) == 0:
            cls._instances[None] = [super().__new__(cls)]
        return cls._instances[None][0]

    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except NECSTParameterNameError as e:
            # Translate to specific error
            raise NECSTConfigurationError from e

    @classmethod
    def from_file(cls, __path: Union[os.PathLike, str, IO]):
        try:
            return super().from_file(__path)
        except ParseError as e:
            # Translate to specific error
            raise NECSTConfigurationError from e

    @property
    def _dotnecst(self) -> str:
        from urllib.parse import urlparse

        try:
            url = urlparse(self._metadata.get("path", ""))
            scheme = f"{url.scheme}://" if url.scheme else ""
            path = url.path
            path = path.decode("utf-8") if isinstance(path, bytes) else path
            return scheme + url.netloc + str(Path(path).parent)
        except Exception:
            return ""

    def reload(self):
        """Reload the configuration.

        This method is to reflect the changes made to the configuration file or the
        change of environment variables.

        """
        path = find_config()
        original_prefix = self._prefix
        new = self.__class__.from_file(path)
        try:
            new = new.get(original_prefix)
        except KeyError:
            new = self.__class__()

        slots = chain.from_iterable(
            getattr(cls, "__slots__", []) for cls in self.__class__.__mro__
        )
        for attr in slots:
            setattr(self, attr, getattr(new, attr))

        for key, parser in parsers.items():
            try:
                self.attach_parsers(**{key: parser})
            except NECSTParameterNameError:
                pass

        for subcls in self.__class__.__subclasses__():
            for children in subcls._instances.values():
                [child.reload() for child in children]

    @classmethod
    def configure(cls):
        """Copy default configuration files into your local file system.

        Important
        ---------
        The files would be placed in ``$HOME/.necst/``. The file contents MUST BE
        UPDATED accordingly, before using this software to actually drive your
        telescope.

        Caution
        -------
        Running this software with misconfigured parameter files could result in
        catastrophic damage to your telescope, by many possible cause. The consequence
        can include: 1) the antenna becomes uncontrollable because the commands are
        calculated in positive feedback, 2) damage your receivers with applying too
        large voltage, and so on.

        """
        DefaultNECSTRoot.mkdir(exist_ok=True)
        for file in DefaultsPath.glob("*.toml"):
            _target_path = DefaultNECSTRoot / file.name
            if _target_path.exists():
                logger.error(f"'{_target_path}' already exists, skipping...")
                continue
            shutil.copyfile(file, _target_path)
        cls().reload()

    @property
    def parameters(self) -> Dict[str, Any]:
        unwrapped = {}
        for k, v in self._parameters.items():
            parsed = v.parsed
            if isinstance(parsed, Path):
                if urlparse(self._dotnecst).scheme:
                    unwrapped[k] = os.path.join(self._dotnecst, parsed)
                else:
                    unwrapped[k] = Path(self._dotnecst) / parsed
            else:
                unwrapped[k] = parsed
        return unwrapped

    def keys(self) -> KeysView:
        prefix_length = len(self._prefix) + 1 if self._prefix else 0
        return KeysView({k[prefix_length:]: v for k, v in self.parameters.items()})

    def values(self) -> ValuesView:
        return self.parameters.values()

    def items(self) -> ItemsView:
        prefix_length = len(self._prefix) + 1 if self._prefix else 0
        return ItemsView({k[prefix_length:]: v for k, v in self.parameters.items()})

    def __getitem__(self, key: str, /):
        item = super().__getitem__(key)
        if isinstance(item, Path):
            if urlparse(self._dotnecst).scheme:
                item = os.path.join(self._dotnecst, item)
            else:
                item = Path(self._dotnecst) / item
        return item

    def __getattr__(self, key: str, /):
        attr = super().__getattr__(key)
        if isinstance(attr, Path):
            if urlparse(self._dotnecst).scheme:
                attr = os.path.join(self._dotnecst, attr)
            else:
                attr = Path(self._dotnecst) / attr
        return attr

    def __add__(self, other: Any, /):
        if other is None:
            return self
        if not isinstance(other, self.__class__):
            return NotImplemented

        if self._prefix == other._prefix:
            parameters = {k: v.parsed for k, v in self._parameters.items()}
            for k, v in other._parameters.items():
                if (k in parameters) and (parameters[k] != v):
                    raise KeyError(
                        f"Conflicting values for {k!r} found: {parameters[k]} and {v}."
                    )
                parameters[k] = v
            parameters.update(other._parameters)
            inst = self.__class__(self._prefix, **parameters)
            return inst

        length = len(self._prefix) + 1
        prefix_stripped = {
            k[length:]: v.parsed
            for k, v in self._parameters.items()
            if k.startswith(self._prefix)
        }
        length = len(other._prefix) + 1
        prefix_stripped_other = {
            k[length:]: v.parsed
            for k, v in other._parameters.items()
            if k.startswith(other._prefix)
        }
        for k, v in prefix_stripped_other.items():
            if (k in prefix_stripped) and (prefix_stripped[k] != v):
                raise KeyError(
                    f"Conflicting values for {k!r} found:"
                    f" {prefix_stripped[k]!r} and {v!r}."
                )
            prefix_stripped[k] = v
        return self.__class__(**prefix_stripped)

    __radd__ = __add__


def find_config() -> str:
    """Look for the configuration file from environment variables and defaults."""
    root_candidates = [str(DefaultNECSTRoot)]
    specified = environ.necst_root.get()
    if specified is not None:
        root_candidates.insert(0, specified)
    config_file_name = "config.toml"
    candidates = map(lambda x: os.path.join(x, config_file_name), root_candidates)

    for path in candidates:
        try:
            read(path)
            logger.info(f"Importing configuration from {path!r}")
            return path
        except FileNotFoundError:
            logger.debug(f"Config file not found at {path!r}")

    logger.error(
        "Config file not found, using the default parameters. "
        "To create the file with default parameters, run `neclib.configure()`."
    )
    config_path = str(DefaultsPath / "config.toml")
    return str(config_path)


class ConfigurationView(Configuration):
    """Sliced configuration."""

    __slots__ = ()

    _instances: Dict[str, List["ConfigurationView"]] = defaultdict(list)

    def __new__(cls, prefix: str = "", /, **kwargs):
        if prefix == "":
            # Addition of config objects can give non-prefixed object. Contents of such
            # objects are not unique and thus difficult to track the changes. Current
            # implementation detaches such object from `reload` chain by not adding them
            # in `cls._instances`.
            return object().__new__(cls)
        if prefix not in cls._instances:
            new = object().__new__(cls)
            cls._instances[prefix].append(new)
            return new
        return cls._instances[prefix][0]

    def reload(self) -> None:
        try:
            filtered = self._metadata["parent"].get(self._prefix)
            self._parameters = filtered._parameters
            self._aliases = filtered._aliases
        except KeyError:
            self._metadata.pop("parent", None)


Configuration._view_class = ConfigurationView

config = Configuration()
"""Collection of system-wide configurations."""
config.reload()

configure = Configuration.configure



import ogameasure

from ...core.security import busy
from .ccd_controller_base import CCDController


class M100(CCDController):

    """ccd camera, which can capture stars when optical pointing.

    Notes
    -----
    You need to install ``libgphoto2`` library in advance.

    Configuration items for this device:

    pic_captured_path : str
        Save directory path of pictures captured by ccd camera.
        This path must be absolute path.
        e.g. ``/home/pi/data/optical_pointing``

    See defaults setting file in ``neclib/defaults/config.toml``.

    """

    Model = "M100"
    Manufacturer = "Canon"

    Identifier = "host"

    def __init__(self) -> None:
        self.ccd = ogameasure.Canon.m100()

    def capture(self, savepath: str) -> None:
        with busy(self, "busy"):
            self.ccd.capture(savepath)
            return None

    def finalize(self) -> None:
        self.ccd.com.close()



import time
from typing import Dict, List, Tuple

from ...core.math import Random
from .spectrometer_base import Spectrometer


class SpectrometerSimulator(Spectrometer):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        _rand = Random(limits=(0, 1e13)).walk(1e10, 1e2, -10)
        initial = [next(_rand) for _ in range(2**15)]
        self._rand = Random().walk(1e10, 1, -1, initial=initial)

    def get_spectra(self) -> Tuple[float, Dict[int, List[float]]]:
        """Timestamp and dict of spectral data for all boards."""
        return time.time(), {0: next(self._rand).tolist()}

    def finalize(self) -> None:
        pass



"""Collection of data type definitions this package interprets."""

from .ordinal import Ordinal  # noqa: F401
from .parameters import Parameters  # noqa: F401
from .rich_parameters import RichParameters  # noqa: F401
from .status_manager import StatusManager  # noqa: F401
from .value_range import ValueRange  # noqa: F401



"""Informative exception, regarding the NECST system status."""

__all__ = [
    "NECSTAuthorityError",
    "NECSTConfigurationError",
    "NECSTTimeoutError",
    "NECSTAccessibilityWarning",
    "NECSTParameterNameError",
]


class NECSTAuthorityError(PermissionError):
    """Error related to controlling priority.

    Examples
    --------
    >>> raise neclib.NECSTAuthorityError("reason or advise to fix this error")

    """

    pass


class NECSTConfigurationError(Exception):
    """Error related to parameter configuration."""

    pass


class NECSTTimeoutError(TimeoutError):
    """Error related to communication timeout."""

    pass


class NECSTParameterNameError(AttributeError):
    """Error related to limitation on parameter name."""

    pass


class NECSTAccessibilityWarning(UserWarning):
    """Warning on limited usage."""

    pass


class NotInitializedError(TypeError):
    """Error related to object initialization status."""

    pass



"""Collection of type converters which allow type-flexible read-in."""

from .array import NPArrayValidator  # noqa: F401
from .astropy import QuantityValidator, get_quantity  # noqa: F401
from .function_defaults import partial  # noqa: F401



"""Definition of environment variable names recognized by NECST."""

__all__ = [
    "necst_root",
    "ros2_ws",
    "domain_id",
    "record_root",
    "debug_mode",
    "log_level",
]

import os
from dataclasses import dataclass
from typing import Optional


@dataclass
class EnvironmentVariable:
    name: str
    """Name of the environment variable."""
    default: Optional[str] = None
    """Default value of the environment variable."""

    def get(self) -> Optional[str]:
        """Get the value of the environment variable."""
        return os.environ.get(self.name, self.default)


necst_root = EnvironmentVariable("NECST_ROOT")
ros2_ws = EnvironmentVariable("ROS2_WS")
domain_id = EnvironmentVariable("ROS_DOMAIN_ID")
record_root = EnvironmentVariable("NECST_RECORD_ROOT")
debug_mode = EnvironmentVariable("NECST_DEBUG_MODE")
log_level = EnvironmentVariable("NECST_LOG_LEVEL", "INFO")



import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Generator, Iterator, Optional, Tuple, Union

import astropy.units as u
import matplotlib.pyplot as plt
import pandas as pd
from astropy.coordinates import SkyCoord

from ...core import Parameters
from ...core.formatting import html_repr_of_observation_spec
from ...core.types import CoordFrameType, CoordinateType, UnitType
from ...core.units import scan_to_points


class ObservationMode(Enum):
    """Type of observation to be performed at certain coordinate."""

    DRIVE = "#777"
    ON = "#0F5"
    OFF = "#0DF"
    HOT = "#F50"
    SKY = "#0DF"


class TimeKeeper:
    """Judge whether it's time to run constant interval observation or not."""

    def __init__(self, interval: u.Quantity, points_per_scan: int = 1):
        self.interval = interval
        self.last = None
        if self._time_based:
            self.count = 0
        else:
            self.count = u.Quantity(0, self.interval.unit)
        self.points_per_scan = points_per_scan

    @property
    def should_observe(self) -> bool:
        """Return ``True`` if it's time to run observation, otherwise ``False``."""
        if self._time_based:
            self.count = time.time() * u.s
        if self.last is None:
            return True
        return bool((self.count - self.last) >= self.interval)

    @property
    def _time_based(self) -> bool:
        return self.interval.unit.is_equivalent(u.s)

    def increment(self, unit: UnitType) -> None:
        """Increment the counter, which will be compared with ``interval``."""
        if not self._time_based:
            _unity = u.Quantity(1, unit or self.interval.unit)
            unity = _unity.to(
                self.interval.unit, equivalencies=scan_to_points(self.points_per_scan)
            )
            self.count = self.count + unity

    def tell_observed(self) -> None:
        """Tell the time keeper that the observation has been completed."""
        self.last = time.time() * u.s if self._time_based else self.count


@dataclass
class Waypoint:
    mode: ObservationMode
    """Observation mode."""
    target: Optional[Union[str, CoordinateType]] = None
    """Coordinate or name of the target object."""
    reference: Optional[Union[str, CoordinateType]] = None
    """Coordinate or name of the object the other coordinate is defined relative to."""
    scan_frame: Optional[CoordFrameType] = None
    """Coordinate frame in which the scan is performed."""
    start: Optional[Tuple[u.Quantity, u.Quantity]] = None
    """Start position of the scan. The frame is defined in ``scan_frame``."""
    stop: Optional[Tuple[u.Quantity, u.Quantity]] = None
    """Stop position of the scan. The frame is defined in ``scan_frame``."""
    offset: Optional[CoordinateType] = None
    """Offset applied to all the coordinates specified above."""
    speed: Optional[u.Quantity] = None
    """Scan speed."""
    integration: Optional[u.Quantity] = None
    """Integration time."""
    id: Any = None
    """Identifier of the coordinate, if necessary."""

    @property
    def is_scan(self) -> bool:
        """Whether this coordinate contains enough information to perform a scan."""
        params = (self.start, self.stop, self.speed, self.scan_frame)
        return all(x is not None for x in params)

    @property
    def with_offset(self) -> bool:
        """Whether this coordinate should be offset by certain amount."""
        return self.offset is not None

    @property
    def name_query(self) -> bool:
        """Whether the target is specified by its name."""
        return isinstance(self.target, str) or isinstance(self.reference, str)

    @property
    def coordinates(self) -> SkyCoord:
        """Return a list of coordinates this section represents.

        Warning
        -------
        This property doesn't perform accurate coordinate calculation when coordinates
        in AltAz frame is involved. Please use :mod:`neclib.coordinates.convert` module
        for accurate coordinate calculation.

        """
        import numpy as np
        from astropy.time import Time

        from ...core import config
        from ..convert import CoordCalculator

        if not hasattr(self, "_calc"):
            self._calc = CoordCalculator(config.location)
        now = Time(time.time(), format="unix")

        nowhere = (float("nan") * u.deg, float("nan") * u.deg, "fk5")

        if self.name_query:
            coord = self._calc.name_coordinate(self.target or self.reference, now)
            coord = coord.realize()
        else:
            target = self.target or self.reference or nowhere
            coord = self._calc.coordinate(
                lon=target[0], lat=target[1], frame=target[2], time=now
            )

        if self.with_offset:
            delta = self._calc.coordinate_delta(
                d_lon=self.offset[0], d_lat=self.offset[1], frame=self.offset[2]
            )
            coord = coord.cartesian_offset_by(delta)

        if self.is_scan:
            delta_start = self._calc.coordinate_delta(
                self.start[0], self.start[1], self.scan_frame
            )
            delta_stop = self._calc.coordinate_delta(
                self.stop[0], self.stop[1], self.scan_frame
            )
            start = coord.cartesian_offset_by(delta_start)
            stop = coord.cartesian_offset_by(delta_stop)
            lon = np.r_[start.lon, stop.lon]
            lat = np.r_[start.lat, stop.lat]
            coord = self._calc.coordinate(
                lon=lon, lat=lat, frame=self.scan_frame, time=now
            )

        if coord.time is None:
            coord.time = time.time()

        return coord.skycoord


class ObservationSpec(Parameters, ABC):
    __slots__ = ("_executing",)
    _repr_frame: CoordFrameType = "fk5"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._executing: Optional[Generator[Waypoint, None, None]] = None

    @abstractmethod
    def observe(self) -> Generator[Waypoint, None, None]:
        """Define all the observation steps which will be interpreted and commanded."""
        ...

    def __next__(self) -> Waypoint:
        if self._executing is None:
            self._executing = self.observe()

        # If generator created above contain nothing, raise StopIteration
        if self._executing is None:
            raise StopIteration
        return next(self._executing)

    def __iter__(self) -> Iterator[Waypoint]:
        return self

    def _repr_html_(self) -> str:
        return super()._repr_html_() + html_repr_of_observation_spec(
            self, frame=self._repr_frame
        )

    def hot(self, id: Any = None, integration: Optional[u.Quantity] = None):
        # Infer the duration by general keyword
        if integration is None:
            integration = self["integ_hot"]
        return Waypoint(mode=ObservationMode.HOT, id=id, integration=integration)

    def off(
        self,
        id: Any = None,
        integration: Optional[u.Quantity] = None,
        *,
        relative: Optional[bool] = None,
        on_coord: Optional[CoordinateType] = None,
        off_coord: Optional[CoordinateType] = None,
    ) -> Waypoint:
        # Infer the duration by general keyword
        if integration is None:
            integration = self["integ_off"]
        if relative is None:
            relative = self["relative"]
        if on_coord is None:
            on_coord = self._reference
        if off_coord is None:
            if relative:
                lon, lat = self["delta_lambda"], self["delta_beta"]
                off_coord = (lon, lat, self["coord_sys"])
            else:
                off_coord = (self["lambda_off"], self["beta_off"], self["coord_sys"])

        kwargs = dict(mode=ObservationMode.OFF, id=id, integration=integration)
        if relative:
            kwargs.update(reference=on_coord, offset=off_coord)
        else:
            kwargs.update(target=off_coord)
        return Waypoint(**kwargs)

    @property
    def coords(self) -> pd.DataFrame:
        """Crudely calculated waypoints this object represents.

        Warning
        -------
        This property doesn't perform accurate coordinate calculation when coordinates
        in AltAz frame is involved.

        """
        waypoints = self.observe()
        waypoint_summary = []

        last_coord = None
        for i, wp in enumerate(waypoints):
            coord = wp.coordinates.transform_to(self._repr_frame)
            lon = coord.data.lon << u.deg
            lat = coord.data.lat << u.deg
            try:
                nan_coord = (lon != lon) or (lat != lat)
            except ValueError:
                nan_coord = (lon != lon).any() or (lat != lat).any()

            transition_lon, transition_lat = [], []
            if (coord.size == 0) or nan_coord:
                pass  # Cannot determine where the observation will be taken place.
            elif coord.size == 1:
                if last_coord is not None:
                    transition_lon = [last_coord[0], lon]
                    transition_lat = [last_coord[1], lat]
                last_coord = (lon, lat)
            else:
                if last_coord is not None:
                    transition_lon = [last_coord[0], lon[0]]
                    transition_lat = [last_coord[1], lat[0]]
                last_coord = (lon[-1], lat[-1])

            coord = [
                dict(lon=_lon, lat=_lat, mode=ObservationMode.DRIVE)
                for _lon, _lat in zip(transition_lon, transition_lat)
            ]
            try:
                _coord = [
                    dict(lon=_lon, lat=_lat, mode=wp.mode)
                    for _lon, _lat in zip(lon, lat)
                ]
                coord.extend(_coord)
            except TypeError:
                coord.extend([dict(lon=lon, lat=lat, mode=wp.mode)])

            for c in coord:
                drive = c["mode"] == ObservationMode.DRIVE
                c.update(
                    waypoint_index=i,
                    scan_frame=None if drive else wp.scan_frame,
                    speed=None if drive else wp.speed,
                    integration=None if drive else wp.integration,
                    id=None if drive else wp.id,
                )
            waypoint_summary.extend(coord)
        return pd.DataFrame(waypoint_summary).set_index("waypoint_index")

    @property
    def fig(self) -> plt.Figure:
        """Figure of crudely calculated telescope driving path.

        Notes
        -----
        If you need ``Axes`` object, use ``fig.axes`` attribute.

        """
        waypoints = self.coords

        with plt.style.context("dark_background"), plt.rc_context(
            {"font.family": "serif", "font.size": 9}
        ):
            fig, ax = plt.subplots(figsize=(3, 3), dpi=150)
            ax.set(
                xlabel="Longitude [deg]",
                ylabel="Latitude [deg]",
                title=f"Waypoints in {self._repr_frame.upper()} frame",
                aspect=1,
            )
            if self._repr_frame.lower() not in ["altaz", "horizontal", "azel"]:
                # Celestial coordinate frames are generally right-handed.
                ax.invert_xaxis()
            for _, _coords in waypoints.groupby(level=0):
                for mode, coords in _coords.groupby("mode", sort=False):
                    lon = coords["lon"] << u.deg
                    lat = coords["lat"] << u.deg
                    try:
                        nan_coord = (lon != lon) or (lat != lat)
                    except ValueError:
                        nan_coord = (lon != lon).any() or (lat != lat).any()

                    if nan_coord:
                        pass  # Cannot determine where to plot.
                    elif len(coords) == 1:
                        ax.plot(lon, lat, ".", c=mode.value, ms=5, alpha=0.9)
                    else:
                        ax.plot(
                            lon,
                            lat,
                            c=mode.value,
                            lw=0.5,
                            ms=1,
                            alpha=0.9,
                            zorder=-1 if mode == ObservationMode.DRIVE else 1,
                        )

            ax.grid(True, c="#767")
            fig.tight_layout()
            plt.close(fig)
            return fig

    @property
    def _reference(self) -> Union[str, Tuple[float, float, str]]:
        if None in (self["lambda_on"], self["beta_on"]):
            return self["target"]
        else:
            return (self["lambda_on"], self["beta_on"], self["coord_sys"])



"""Interface to access NECLIB parameters.

This module defines the NECLIB standard parameter interface. The values can be any
Python built-in type, or physical quantity which need units specified.

The parameters would be stored in TOML files, in the following format:

.. code-block:: toml

   [parameter_kind]
   parameter_name = value

   [another_parameter_kind]
   "parameter_with_units[units]" = value
   parameter_without_units = value

The parameters should be grouped by their kind, which is specified by the TOML table
name. This kind won't be kept in the parameter interface, but we require its use to
keep the parameter files readable.

"""

import os
import re
import warnings
from itertools import chain
from typing import IO, Any, Dict, List, Tuple, Union

import astropy.units as u
from astropy.coordinates import Angle

from ..exceptions import NECSTAccessibilityWarning, NECSTParameterNameError
from ..files import toml
from ..formatting.dict_to_html import html_repr_of_dict


class Parameters:
    """General format of NECLIB parameters.

    This class provides a convenient interface to access the parameters, whose values
    can be physical quantities; which contain units. The parameters can be stored in
    TOML files, so this class also provides a its parser.

    Examples
    --------
    If the parameters are stored in a TOML file, use ``from_file`` method:

    >>> from neclib.parameters import Parameters
    >>> params = Parameters.from_file("parameters.toml")
    >>> params["distance"]
    <Quantity 10. pc>

    You can also provide the parameters as keyword arguments, but physical units may not
    be parsed:

    >>> params = Parameters(distance="10 pc")
    >>> params["distance"]
    "10 pc"

    To parse the units, use the following syntax:

    >>> params = Parameters(**{"distance[pc]": 10})
    >>> params["distance"]
    <Quantity 10. pc>

    Of course you can provide many parameters at once:

    >>> params = Parameters(**{"distance[pc]": 10, "inclination[deg]": 45})
    >>> params["distance"]
    <Quantity 10. pc>

    And you can convert the units on class initialization:

    >>> params = Parameters(**{"distance[pc]": "10m")
    >>> params["distance"]
    <Quantity 3.24077929e-16 pc>

    You can also access the parameters as attributes:

    >>> params.distance
    <Quantity 10. pc>

    Notes
    -----
    The units are parsed using ``astropy.units`` and ``astropy.coordinates.Angle``. The
    units are parsed from the parameter name, so the parameter name must be in the
    following format: ``parameter_name[unit]``.

    """

    __slots__ = ("_parameters", "_metadata", "_aliases")

    _unit_matcher = re.compile(r"(\w*)\[([\w/\s\*\^-]*)\]")
    # Matches `key[unit]` format. Key should be composed of alphanumeric characters
    # and/or underscores. Unit can be composed of alphanumeric characters, whitespaces
    # and `*/^-`.

    _angle_units: List[str] = [unit.name for unit in u.rad.find_equivalent_units()] + [
        alias for unit in u.rad.find_equivalent_units() for alias in unit.aliases
    ]

    def __init__(self, **kwargs) -> None:
        self._parameters = dict(map(self._parse, kwargs.keys(), kwargs.values()))
        self._metadata = {}
        self._aliases: Dict[str, str] = {}

    @classmethod
    def from_file(cls, file: Union[os.PathLike, str, IO], /):
        """Read parameters from a TOML file.

        Parameters
        ----------
        file
            The file path or file object.

        """
        file_path = isinstance(file, (os.PathLike, str))
        params = toml.read(file)
        params = {k: v for x in params.unwrap().values() for k, v in x.items()}

        inst = cls(**params)
        if file_path:
            inst._metadata["path"] = file
        return inst

    def attach_aliases(self, **kwargs: str) -> None:
        """Attach aliases to the parameters.

        Different names can be used to access the same parameter. This is useful when
        the parameter name is too long, or when the parameter name has some dialects.

        Parameters
        ----------
        **kwargs
            The aliases to attach. The keys are the alias names, and the values are the
            existing parameter names.

        """
        for k, v in kwargs.items():
            self._validate(k)
            if v not in self._parameters:
                raise NECSTParameterNameError(f"Unknown parameter: {v!r}")
            if k in self._parameters:
                raise NECSTParameterNameError(
                    f"Cannot create alias with existing parameter name: {k!r}"
                )
            if k not in self._aliases:
                self._aliases[k] = v

    def __str__(self) -> str:
        width = max(len(x) for x in self._parameters)
        params = [
            f"{k:{width}s} = {v} ({type(v).__name__})"
            for k, v in self._parameters.items()
        ]
        return self.__class__.__name__ + "\n" + "\n".join(params)

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({len(self._parameters)} elements)"

    def _repr_html_(self) -> str:
        return html_repr_of_dict(
            self._parameters,
            type(self),
            aliases=self._aliases,
            metadata=self._metadata,
        )

    def _validate(self, key: str, /) -> None:
        """Check the parameter name is not reserved.

        Reserved names are `__slots__` contents (possible instance variables) and
        `__class__.__dict__` contents (class variables, including method names).

        Parameters
        ----------
        key
            The parameter name to check.

        """
        slots = chain.from_iterable(
            getattr(cls, "__slots__", []) for cls in self.__class__.__mro__
        )
        if (key in self.__class__.__dict__) or (key in slots):
            raise NECSTParameterNameError(f"Reserved name: {key!r}")
        if not key.isidentifier():
            warnings.warn(
                f"{key!r} isn't a valid Python identifier, "
                "so attribute access isn't available.",
                NECSTAccessibilityWarning,
            )

    def _parse(self, k: str, v: Any, /) -> Tuple[str, Any]:
        if v == {}:
            # Use empty mapping as empty value. This rule is not compliant with TOML
            # spec, but we need a valid TOML value which represents empty.
            v = None

        unit_match = self._unit_matcher.match(k)
        if unit_match is None:
            # When no unit is specified, the value is used as is.
            self._validate(k)
            return k, v

        key, unit = unit_match.groups()
        self._validate(key)
        if v is None:
            return key, v
        if unit in self._angle_units:
            return key, Angle(v, unit=unit)
        return key, u.Quantity(v, unit=unit)

    def __getitem__(self, key: str, /) -> Any:
        if key in self._parameters:
            return self._parameters[key]
        if key in self._aliases:
            return self._parameters[self._aliases[key]]

    def __getattr__(self, key: str, /) -> Any:
        if key in self._parameters:
            return self._parameters[key]
        if key in self._aliases:
            return self._parameters[self._aliases[key]]

    @property
    def parameters(self) -> Dict[str, Any]:
        """Return a copy of the raw parameters."""
        return self._parameters.copy()

    def __eq__(self, other: Any, /) -> bool:
        if not isinstance(other, Parameters):
            return NotImplemented
        return self._parameters == other._parameters

    def __lt__(self, other: Any, /) -> bool:
        if not isinstance(other, Parameters):
            return NotImplemented
        if not set(self._parameters) < set(other._parameters):
            return False
        for k, v in self._parameters.items():
            if v != other._parameters[k]:
                return False
        return True

    def __gt__(self, other: Any, /) -> bool:
        if not isinstance(other, Parameters):
            return NotImplemented
        if not set(self._parameters) > set(other._parameters):
            return False
        for k, v in other._parameters.items():
            if v != self._parameters[k]:
                return False
        return True

    def __le__(self, other: Any, /) -> bool:
        eq, lt = self.__eq__(other), self.__lt__(other)
        if (eq is NotImplemented) or (lt is NotImplemented):
            return NotImplemented
        return eq or lt

    def __ge__(self, other: Any, /) -> bool:
        eq, gt = self.__eq__(other), self.__gt__(other)
        if (eq is NotImplemented) or (gt is NotImplemented):
            return NotImplemented
        return eq or gt

    def __ne__(self, other: Any, /) -> bool:
        eq = self.__eq__(other)
        return NotImplemented if eq is NotImplemented else not self.__eq__(other)



from abc import abstractmethod

from ..device_base import DeviceBase


class CCDController(DeviceBase):
    @abstractmethod
    def capture(self) -> None:
        ...

    @abstractmethod
    def finalize(self) -> None:
        ...



"""Drive path calculators.

Notes
-----
The modules in this namespace should not inherit from ``CoordCalculator`` or its
subclasses, since such dependency would cause circular imports. Instead, the calculator
should be passed as an argument to the path calculator's constructor.

"""

from .linear import Accelerate, Linear, Standby  # noqa: F401
from .path_base import ControlContext, Index  # noqa: F401
from .track import Track  # noqa: F401



__all__ = ["describe_frame", "parse_frame"]

import re
from dataclasses import dataclass
from functools import lru_cache
from typing import Callable, ClassVar, Dict, Type, Union

from astropy.coordinates import (
    Angle,
    BaseCoordinateFrame,
    SkyOffsetFrame,
    frame_transform_graph,
)


@dataclass
class Frame:
    """Converts between frame objects and their string representations."""

    frame: Union[BaseCoordinateFrame, Type[BaseCoordinateFrame]]
    """Frame object."""

    aliases: ClassVar[Callable[[], Dict[str, str]]] = lambda: {
        "j2000": "fk5",
        "b1950": "fk4",
        "horizontal": "altaz",
    }
    """A dictionary of coordinate frame name dialects."""
    # Since dictionary is mutable, this alias list is defined as a callable.

    @staticmethod
    def _parse(frame: str) -> Union[BaseCoordinateFrame, Type[BaseCoordinateFrame]]:
        """Convert a string representation into a frame object."""
        # Search for the frame in AstroPy's built-in frames
        _parsed = frame_transform_graph.lookup_name(frame.lower())
        if _parsed is not None:
            return _parsed

        # Parse SkyOffsetFrame specs
        _parsed = re.match(
            r"origin=([a-z0-9]*)\(-?([a-z\d\.]*),-?([a-z\d\.]*)\),"
            r"rotation=(-?[a-z\d\.]*)",
            frame.lower().replace(" ", ""),
        )
        if _parsed is not None:
            base_frame, lon, lat, rotation = _parsed.groups()
            BaseFrame = frame_transform_graph.lookup_name(base_frame.lower())
            rotation = Angle(rotation)
            if BaseFrame is None:
                raise ValueError(f"Unknown frame {base_frame!r}")
            return SkyOffsetFrame(origin=BaseFrame(lon, lat), rotation=rotation)
        raise ValueError(f"Could not parse frame {frame!r}")

    @staticmethod
    def _describe(frame: Union[BaseCoordinateFrame, Type[BaseCoordinateFrame]]) -> str:
        """Convert a frame object into a string representation."""
        # Built-in frames
        if not isinstance(frame, SkyOffsetFrame):
            return frame.name  # type: ignore

        # Get string representation of SkyOffsetFrame
        center: BaseCoordinateFrame = frame.origin  # type: ignore
        attr_names = center.representation_component_names
        attr_name_mapping = {v: k for k, v in attr_names.items()}
        lon = getattr(center, attr_name_mapping["lon"])
        lat = getattr(center, attr_name_mapping["lat"])

        origin = f"{center.name}({lon}, {lat})"
        rotation = f"{frame.rotation}"
        return f"origin={origin}, rotation={rotation}"

    @classmethod
    @lru_cache(maxsize=16)
    def from_string(cls, frame: str, /) -> "Frame":
        """Create Frame object, parsing string representation."""
        frame = frame.lower()
        for k, v in cls.aliases().items():
            frame = frame.replace(k, v)
        parsed_frame = cls._parse(frame)
        return cls(parsed_frame)

    def __str__(self) -> str:
        return self._describe(self.frame)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._describe(self.frame)})"


def parse_frame(frame: str, /) -> Union[Type[BaseCoordinateFrame], BaseCoordinateFrame]:
    """Parse a frame string and create a frame object.

    Parameters
    ----------
    frame
        String representation of coordinate frame.

    Returns
    -------
    Parsed frame object.

    Examples
    --------
    >>> neclib.coordinates.parse_frame("fk5")
    <class 'astropy.coordinates.builtin_frames.fk5.FK5'>
    >>> neclib.coordinates.parse_frame("altaz")
    <class 'astropy.coordinates.builtin_frames.altaz.AltAz'>
    >>> neclib.coordinates.parse_frame("origin=FK5(10deg, 0deg), rotation=20deg")
    <SkyOffsetFK5 Frame (equinox=J2000.000, rotation=20.0 deg, origin=<FK5 Coordinate
    (equinox=J2000.000): (ra, dec) in deg
    (10., 0.)>)>

    """
    return Frame.from_string(frame).frame


def describe_frame(
    frame: Union[BaseCoordinateFrame, Type[BaseCoordinateFrame]], /
) -> str:
    """Get string representation of a frame.

    Parameters
    ----------
    frame
        Frame object to describe.

    Returns
    -------
    String representation of the frame.

    Examples
    --------
    >>> neclib.coordinates.describe_frame(FK5)
    'fk5'
    >>> neclib.coordinates.describe_frame(AltAz())
    'altaz'
    >>> neclib.coordinates.describe_frame(
    ...     SkyOffsetFrame(origin=FK5("10deg", "0deg"), rotation="20deg")
    ... )
    'origin=fk5(10deg, 0deg), rotation=20deg'

    """
    return str(Frame(frame))



import astropy.units as u
import ogameasure

from ... import utils
from ...core.security import busy
from .attenuator_base import Attenuator


class RHIO10(Attenuator):
    """Attenuator, which can attennuate IF sigal power.

    Notes
    -----

    Configuration items for this device:

    host : str
        IP address for ethernet communicator.

    port : int
        ethernet port of using devices.

    channel : Dict[str, int]
        Human-readable channel name. The value should be
        mapping from human readableversion (str) to
        device level identifier (int). You can assign any name to the
        channels up to two channels. For example: `{ CH1 = 1, CH2 = 2}`

    """

    Manufacturer: str = "SENA"
    Model: str = "RHIO10"

    Identifier = "host"

    @utils.skip_on_simulator
    def __init__(self) -> None:
        com = ogameasure.ethernet(self.Config.host, self.Config.port)
        self.io = ogameasure.SENA.adios(com)

    def get_loss(self, id: str) -> u.Quantity:
        with busy(self, "busy"):
            ch = self.Config.channel[id]
            try:
                if ch == 1:
                    return self.io.get_att1() << u.dB
                elif ch == 2:
                    return self.io.get_att2() << u.dB
            except IndexError:
                pass
            raise ValueError(f"Invalid channel: {ch}")

    def set_loss(self, dB: int, id: str) -> None:
        with busy(self, "busy"):
            ch = self.Config.channel[id]
            self.io._set_att(ch, int(dB))

    def finalize(self) -> None:
        self.io.com.close()
